from time import time
from json import loads, dumps
from redis import ConnectionPool, StrictRedis
import logging
import os


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SetRedis:
    def __init__(self, host=None, port=6379, db=0, decode_responses=True, **kwargs):
        """
        :param kwargs: redis config args.
        """
        self.__conf_redis = {**{'host': host, 'port': port, 'db': db},
                             **kwargs} if host else {}
        self.redis = StrictRedis(connection_pool=ConnectionPool(**{**self.connetion(), **self.__conf_redis}),
                                 decode_responses=decode_responses)

    def connetion(self):
        """
        get redis config form your environ or enter.
        redis config name use >CONF_FOR_REDISCONF<
        :return: dict of redis config
        """
        try:
            redis_conf = self.__conf_redis or eval(os.environ['CONF_FOR_REDISCONF'])
        except KeyError:
            os.environ['CONF_FOR_REDISCONF'] = input(
                "WARNING:\n Haven't catch redis config from your env.\n Entry CONF_FOR_REDISCONF args：")
            redis_conf = eval(os.environ['CONF_FOR_REDISCONF'])
        return redis_conf

    def getkeys(self):
        """
        get all fields from redis.
        :return: get fields(keys) list
        """
        return self.redis.keys()

    def getconf(self, name):
        """
        get value according fields.
        :param name: redis fields
        :return: get keys and values according fields.
        """
        res = self.redis.zrevrange(name, 0, 0)
        if res:
            return loads(self.redis.zrevrange(name, 0, 0)[0])
        else:
            return {}

    def set_conf(self, name, mapping=None, notify=True, **kwargs):
        """
        create or update mapping according fields.
        :param notify: show log notice.
        :param name: redis key
        :param mapping: <dict> data what is create or update
        """
        mapping = {**mapping, **kwargs} if mapping else kwargs
        if mapping:
            mapping = {dumps(mapping): int(time())}
            self.redis.zadd(name, mapping)
            logger.info(f"→{name}← has been setup the mapping: →{mapping}← done.") if notify else None



from configredis import SetConfig, Tools, SetRedis


con = SetConfig.ConfigArgs()

SetConfig.defaultconfig(
    c_env='default',
    disk_name='TenD'
)

SetConfig.devconfig(
    c_env='dev',
    celery_broker="amqp://user:password@172.0.0.1:5672//",
    r_flow={'host': Tools.host_ip('httpbin.org'), 'port': 7000, 'db': 0, "cluster": True},
    r_flow1={'host': 'httpbin.org', 'port': 7000, 'db': 0, "cluster": True}
)

SetConfig.proconfig(
    c_env='pro',
    celery_broker="amqp://user:password@172.0.0.1:5672//",
    disk_name="TenB"
)

SetConfig.envconfig(
    'custom_env',
    c_env='custom_env'
)

config = SetConfig.configs('bin.org')  # 如果明文配置env，会以明文配置优先，如果配置文件删除明文配置，则会查找redis最新配置，SetConfig.configs会自动把明文配置持久化到redis  # noqa


if __name__ == '__main__':
    print(config)  # 获得当前env的配置
    # print(SetConfig.getconfs())  # 查看redis中项目目前所有版本配置
    # SetConfig.print_config()  # 罗列项目所有版本的配置信息，并根据index打印相关配置，方便复制到配置文件, 顺序从新到旧
    # print(con['custom_env'])  # 根据env查看redis储存的最新配置
    # SetConfig.upsert_field_to_redis('custom_env', disk_name='TenC')  # 明文配置删除的情况下，给redis中的最新配置增加字段，不要跟明文配置同时使用，会导致redis版本错乱  # noqa
    # print(Tools.lookup_redis_proj_config())  # 显示redis存储的最新配置
    # print(Tools.lookup_proj_config(replace_domain='bin.org'))  # 显示当前所有env配置，明文优先于redis配置
    # print(SetRedis().getkeys())  # 查看redis所有的项目名称
    # SetConfig.delconfs()  # 罗列项目所有版本的配置信息，并根据index删除相关配置, 顺序从旧到新



from configredis.setredis import SetRedis
from json import dumps, loads
from datetime import datetime
import sys
import logging
import os
import socket
import re

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
redis = SetRedis()


class Tools:
    @classmethod
    def host_ip(cls, hostname: str):
        return socket.gethostbyname(hostname)

    @classmethod
    def project_name(cls, path=os.getcwd(), dirs=(".git",), default=None):
        """
        get current project name.
        :param path: scripts path
        :param dirs: save rule e.g: catch git
        :param default: if set default then save to the default absolute path
        :return: project root directory name
        """
        prev, path = None, os.path.abspath(path)
        while prev != path:
            if any(os.path.isdir(os.path.join(path, d)) for d in dirs):
                return path.split('/')[-1]
            prev, path = path, os.path.abspath(os.path.join(path, os.pardir))
        return default

    @classmethod
    def lookup_redis_proj_config(cls, proj_name=None):
        """
        lookup project config on redis.
        :param proj_name: project name which lookup from redis.
        """
        proj_name = proj_name or cls.project_name()
        return redis.getconf(proj_name)

    @classmethod
    def lookup_proj_config(cls, env=None, replace_domain=''):
        """
        lookup current project config.
        """
        conf_mapping_ = SetConfig.local_mapping_ or SetConfig.redis_mapping_
        return SetConfig.with_domain(conf_mapping_, replace_domain, env)


class ConfigUpdate:
    project_name_ = Tools.project_name()
    local_mapping_ = {}
    redis_mapping_ = Tools.lookup_redis_proj_config()

    @classmethod
    def upsert_config_to_redis(cls, mapping=None, notify=True):
        """
        insert or update current config to redis.
        """
        redis.set_conf(cls.project_name_, mapping=mapping or cls.local_mapping_, notify=False)
        logger.info(f" {cls.project_name_}'s config has been written to redis.") if notify else None

    @classmethod
    def upsert_field_to_redis(cls, env='default', notify=True, **kwargs):
        conf_mapping_ = cls.local_mapping_ or cls.redis_mapping_
        conf_mapping_[env] = {**conf_mapping_[env], **kwargs}
        redis.set_conf(cls.project_name_, mapping=conf_mapping_, notify=False)
        logger.info(f" {cls.project_name_}'s {env} config fileds has been written to redis.") if notify else None

    @classmethod
    def with_domain(cls, configs, replace_domain='', env=None):
        configs = dumps(configs)
        conf = loads(cls.__change_host(configs, replace_domain) if replace_domain else configs)
        return {**conf.get('default', {}), **conf.get(env)} if env else conf

    @classmethod
    def __change_host(cls, conf, domain: str):
        pattern = r'[\w\.]+' + domain.replace('.', r'\.')
        conf, pattern = conf, re.compile(r'' + pattern)
        n_d = {i: Tools.host_ip(i) for i in pattern.findall(conf)}
        for k, v in n_d.items():
            conf = conf.replace(k, v)
        return conf

    class ConfigArgs:
        """
        get config args value from redis.
        """

        def __getitem__(self, env):
            """
            sample:
            :param env: name is 'dev'
            :return: dict e.g: {'disk_name': 'TenTBc'}
            """
            return Tools.lookup_redis_proj_config().get(env, {})


class SetConfig(ConfigUpdate):
    __init_conf = ['default', 'dev', 'pro']

    @classmethod
    def defaultconfig(cls, **mapping):
        """
        set default env args.
        need import this func to your config document.
        :param mapping: dict
        """
        cls.envconfig('default', **mapping)

    @classmethod
    def devconfig(cls, **mapping):
        """
        set development env args.
        need import this func to your config document.
        :param mapping: dict
        """
        cls.envconfig('dev', **mapping)

    @classmethod
    def proconfig(cls, **mapping):
        """
        Set production env args.
        need import this func to your config document.
        :param mapping: dict
        """
        cls.envconfig('pro', **mapping)

    @classmethod
    def envconfig(cls, env_: str, **mapping):
        """
        Set production env args.
        need import this func to your config document.
        :param env_: env name
        :param mapping: dict
        """
        cls.local_mapping_[env_] = mapping

    @classmethod
    def configs(cls, replace_domain: str = None, on_redis=True):
        """
        get config info according env args. [defult/dev/pro]
        need import this func to your END of config document.
        :return: env config info as dict.
        """
        env = sys.argv[-1].lower()
        if not cls.local_mapping_:
            logger.warning(' Not catch local project config, will use Redis project config.')
        conf_mapping_ = cls.local_mapping_ or cls.redis_mapping_
        if env not in conf_mapping_:
            logger.warning(' Not catch env args or defultconfig not setup. e.g: python sample.py dev|pro')
            logger.warning(' Will start under dev env? exit: ctrl+c')
            env = 'dev'
        logger.info(f' Current env is {env}\n')
        if on_redis:
            cls.upsert_config_to_redis(cls.local_mapping_, notify=False)
        return cls.with_domain(conf_mapping_, replace_domain, env)

    @classmethod
    def delconfs(cls, proj_name=None):
        """
        delete concent from name or key.
        :param proj_name: proj_name
        """
        proj_name = proj_name or cls.project_name_
        confs, res = cls.getconfs(proj_name, score_cast_func=int), 0
        del_items, choose = cls.__choose_items(confs)
        if '0' in choose:
            res = redis.redis.delete(proj_name)
        elif del_items:
            res = redis.redis.zrem(proj_name, *del_items)
        if res:
            logger.info(' Redis configs has been deleted.')

    @classmethod
    def getconfs(cls, proj_name=None, score_cast_func=None):
        proj_name = proj_name or cls.project_name_
        f_score = score_cast_func or cls.__f_score
        return {str(i[1]): i[0] for i in redis.redis.zscan_iter(proj_name, score_cast_func=f_score)}

    @classmethod
    def print_config(cls, proj_name=None):
        proj_name = proj_name or cls.project_name_
        info = ''
        confs = cls.getconfs(proj_name, score_cast_func=int)
        print_items = cls.__choose_items(confs, stamp=True, reverse=True)[0]
        for i in print_items:
            info += f"\n\n>>{cls.__f_score(i)}<<\n\n" + "\n\n".join(
                [cls.__f_env_key(k) + cls.__f_env_conf(v) for k, v in loads(confs[i]).items()]) + '\n'
        print(f"{info}")

    @classmethod
    def indent(cls, num):
        return ''.rjust(num, ' ')

    @classmethod
    def __choose_items(cls, confs, stamp=False, reverse=False):
        sore_keys = list(confs.keys())
        sore_keys.sort(reverse=reverse)
        p_conf = "\n".join(["0: All", *[f"{k + 1}: {cls.__f_score(v)} -> {confs[v]}"
                                        for k, v in enumerate(sore_keys)]])
        print(p_conf)
        choose = input('Enter index num which config choosed (split with ","):').replace(' ', '').split(',')
        return [sore_keys[i] if stamp else confs[sore_keys[i]] for i in range(len(sore_keys)) if str(i + 1) in choose
                ], choose

    @classmethod
    def __f_env_key(cls, env_key):
        return (f"SetConfig.{env_key}config(\n" if env_key in cls.__init_conf else
                f"SetConfig.envconfig(\n{cls.indent(4)}'{env_key}',\n")

    @classmethod
    def __f_env_conf(cls, env_conf: dict):
        return f'{cls.indent(4)}' + f'{cls.indent(4)}'.join(
            [f'{k}="{v}",\n' if isinstance(v, str) else f'{k}={v},\n' for k, v in env_conf.items()]) + ")"

    @classmethod
    def __f_score(cls, stamp):
        return f"{datetime.fromtimestamp(int(stamp))} ({stamp})"
